from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig

from wbcrm.models.activities import Activity, ActivityParticipant
from wbcrm.synchronization.activity.shortcuts import get_backend

DESCRIPTION: str = _(
    "<p> Are you sure you want to delete all the future instances of this activity? <br> \
    Only 'Planned' and 'Canceled' activities will be deleted. <br>\
    Depending on the number of activities to be deleted <br> \
    it may take some time until the deleted activities are no longer displayed in the activity list. </p>",
)


class ActivityButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self) -> set:
        if self.view.kwargs.get("pk"):
            return {
                bt.WidgetButton(
                    label=_("Parent Activity"), icon=WBIcon.CALENDAR.icon, key="get_parent_occurrence", weight=110
                ),
                bt.ActionButton(
                    method=RequestType.DELETE,
                    identifiers=("wbcrm:activity",),
                    key="delete_next_occurrences",
                    label=_("Delete Next Occurrences"),
                    icon=WBIcon.DELETE.icon,
                    description_fields=DESCRIPTION,
                    title=_("Delete"),
                    action_label=_("Delete"),
                    weight=140,
                ),
                bt.WidgetButton(label=_("Next Activity"), icon=WBIcon.NEXT.icon, key="next_occurrence", weight=130),
                bt.WidgetButton(
                    label=_("Previous Activity"),
                    icon=WBIcon.PREVIOUS.icon,
                    key="previous_occurrence",
                    weight=120,
                ),
            }

        return set()


class ActivityParticipantButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        buttons = set()
        if not self.view.kwargs.get("pk"):
            base_url = reverse("wbcrm:activity-list", args=[], request=self.request)
            activity: Activity = Activity.all_objects.get(id=self.view.kwargs.get("activity_id"))
            if activity.period:
                participants_id_set: set[int] = set(activity.participants.values_list("id", flat=True))
                id_str = ",".join(str(id) for id in participants_id_set)

                start = activity.period.lower.date()
                end = activity.period.upper.date()

                endpoint = f"{base_url}?participants={id_str}&period={start:%Y-%m-%d},{end:%Y-%m-%d}"

                buttons.add(
                    bt.WidgetButton(
                        endpoint=endpoint,
                        label=_("Show Participants' Activities"),
                        icon=WBIcon.CALENDAR.icon,
                    )
                )

            # Activity sync button to send invitation to external participants
            if get_backend():
                if activity.activity_participants.filter(
                    participation_status=ActivityParticipant.ParticipationStatus.PENDING_INVITATION
                ).exists():
                    buttons.add(
                        bt.ActionButton(
                            method=RequestType.PATCH,
                            endpoint=reverse(
                                "wbcrm:activity-participant-send-external-invitation",
                                args=[activity.id],
                                request=self.request,
                            ),
                            label="Send invitation to External",
                            icon=WBIcon.SEND_LATER.icon,
                        )
                    )
        return buttons



import itertools
import json
from unittest.mock import patch

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.test import APIRequestFactory
from wbcore.permissions.registry import user_registry

from wbcrm.factories import ActivityFactory
from wbcrm.models import Activity, ActivityParticipant
from wbcrm.synchronization.activity.shortcuts import get_backend
from wbcrm.typings import User as UserDTO

from .fixtures import MSGraphFixture, TestOutlookSyncFixture

User = get_user_model()


@pytest.mark.django_db
class TestInitController:
    def test_init(self):
        global_preferences_registry.manager()["wbactivity_sync__sync_backend_calendar"] = ""
        controller = get_backend()
        assert controller is None

        global_preferences_registry.manager()["wbactivity_sync__sync_backend_calendar"] = (
            "wbcrm.synchronization.activity.backends.outlook.backend.OutlookSyncBackend"
        )
        controller2 = get_backend()
        assert controller2.backend


@pytest.mark.django_db
class TestController(TestOutlookSyncFixture):
    def setup_method(self):
        global_preferences_registry.manager()["wbactivity_sync__sync_backend_calendar"] = (
            "wbcrm.synchronization.activity.backends.outlook.backend.OutlookSyncBackend"
        )

    @pytest.fixture()
    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def controller(self, mock_msgraph):
        controller = get_backend()
        controller.backend.open()
        controller.backend.msgraph = MSGraphFixture()
        return controller

    @pytest.mark.parametrize("type_request", [None, "validationToken", "admin_consent"])
    def test_handle_inbound_validation_response(self, controller, type_request):
        request1 = APIRequestFactory().get("")
        if type_request:
            request1.GET = request1.GET.copy()
            request1.GET[type_request] = "fake_info"
            assert controller.handle_inbound_validation_response(request1).content.decode("UTF-8") == "fake_info"
        else:
            assert controller.handle_inbound_validation_response(request1) is None

    @pytest.mark.parametrize("client_state", ["secret1", "secret2"])
    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def test_get_events_from_inbound_request(
        self, mock_msgraph, controller, client_state, notification_fixture, teams_event_fixture
    ):
        notification_fixture["client_state"] = client_state
        global_preferences_registry.manager()["wbactivity_sync__outlook_sync_client_state"] = "secret2"
        api_factory = APIRequestFactory()
        request1 = api_factory.post("", data={})
        request2 = api_factory.post(
            "", data=json.dumps({"value": [notification_fixture]}), content_type="application/json"
        )

        mock_msgraph.return_value = controller.backend.msgraph
        controller.backend.msgraph.event = teams_event_fixture
        controller.backend.msgraph.event_by_uid = teams_event_fixture
        events1 = controller.get_events_from_inbound_request(request1)
        events2 = controller.get_events_from_inbound_request(request2)
        assert events1 == []
        if client_state != global_preferences_registry.manager()["wbactivity_sync__outlook_sync_client_state"]:
            assert events1 == events2 == []
        else:
            assert events1 == []
            assert len(events2) == 1
            expected_result = {
                "change_type": notification_fixture["change_type"],
                "resource": notification_fixture["resource"],
                "subscription_id": notification_fixture["subscription_id"],
                "organizer_resource": notification_fixture["resource"],
                **teams_event_fixture,
            }
            assert set(events2[0].keys()) == set(expected_result.keys())

    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def test_user_for_handle_inbound(
        self, mock_msgraph, controller, notification_fixture, teams_event_fixture, user_factory
    ):
        event = {
            "change_type": notification_fixture["change_type"],
            "resource": notification_fixture["resource"],
            "subscription_id": notification_fixture["subscription_id"],
            **teams_event_fixture,
        }
        metadata = {"outlook": {"subscription": {"id": notification_fixture["subscription_id"]}}}
        _, _, user_dto = controller.backend._deserialize(event)
        user_result = UserDTO(metadata=metadata, id=None)
        assert user_dto.id is None
        assert user_dto.metadata == user_result.metadata

        assert controller.get_activity_participant(user_dto) is None
        user = user_factory(is_active=True, is_superuser=True, metadata=metadata)
        assert controller.get_activity_participant(user_dto) == user.profile

    def _get_user_and_activities(self, notification_fixture, user_factory, teams_event_fixture):
        user1 = user_factory(is_active=True)
        user2 = user_factory(
            is_active=True, metadata={"outlook": {"subscription": {"id": notification_fixture["subscription_id"]}}}
        )

        other1_activity = ActivityFactory(creator=user1.profile, participants=(user2.profile,))
        other2_activity = ActivityFactory(
            creator=user1.profile,
            participants=(user2.profile,),
            metadata={"outlook": {"event_id": 1, "event_uid": 2, "resources": ["Users/1/events/1"]}},
        )
        metadata = {
            "outlook": {
                "resources": [notification_fixture["resource"]],
                "organizer_resource": notification_fixture["resource"],
                "event_uid": teams_event_fixture["uid"],
                "event_id": teams_event_fixture["id"],
            }
        }
        activity = ActivityFactory(creator=user1.profile, participants=(user2.profile,))
        Activity.objects.filter(id=activity.id).update(metadata=metadata)
        return {"activities": (other1_activity, other2_activity, activity), "users": (user1, user2)}

    @pytest.mark.parametrize(
        "is_internal_creator, cancel_activity, delete_notification", list(itertools.product([True, False], repeat=3))
    )
    def test_delete_activity(
        self,
        cancel_activity,
        is_internal_creator,
        controller,
        notification_fixture,
        user_factory,
        teams_event_fixture,
        delete_notification,
    ):
        global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_activity"] = cancel_activity

        data = self._get_user_and_activities(notification_fixture, user_factory, teams_event_fixture)
        other1_activity, other2_activity, activity = data["activities"]
        user1, user2 = data["users"]
        if is_internal_creator:
            permission = Permission.objects.get_or_create(
                content_type=ContentType.objects.get_for_model(User), codename="is_internal_user"
            )[0]
            user1.user_permissions.add(permission)
            user_registry.reset_cache()
        activity.refresh_from_db()
        activity_dto = activity._build_dto()
        activity_dto.delete_notification = delete_notification

        user1_dto = UserDTO(metadata=user1.metadata, id=None)
        user2_dto = UserDTO(metadata=user2.metadata, id=None)

        assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}

        controller.delete_activity(activity_dto, user1_dto)
        assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}

        controller.delete_activity(
            activity_dto, user2_dto
        )  # user 2 is not the creator of the activity. it's a participant
        assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}
        if is_internal_creator:
            user2_status = (
                ActivityParticipant.ParticipationStatus.CANCELLED
                if delete_notification
                else ActivityParticipant.ParticipationStatus.PENDING_INVITATION
            )
            assert activity.activity_participants.get(participant=user2.profile).participation_status == user2_status

        controller.delete_activity(
            activity_dto, user1_dto
        )  # user 1 is the creator of the activity but doesn't have an active subscription
        assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}
        user1.metadata = user2.metadata
        user1.save()
        user2.metadata = {}
        user2.metadata = {"outlook": {"subscription": {"id": "new_subscription1"}}}
        user2.save()
        user1.refresh_from_db()
        user1_dto = UserDTO(metadata=user1.metadata, id=None)

        controller.delete_activity(
            activity_dto, user1_dto
        )  # user 1 is the creator of the activity with an active subscription
        if delete_notification and is_internal_creator and not cancel_activity:
            assert set(Activity.objects.all()) == {other1_activity, other2_activity}
        else:
            assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}
        assert set(Activity.all_objects.all()) == {other1_activity, other2_activity, activity}

    @pytest.mark.parametrize(
        "cancel_external_creator_activity, cancel_activity, delete_notification",
        list(itertools.product([True, False], repeat=3)),
    )
    def test_cancel_no_participant_external_creator_activity(
        self,
        cancel_external_creator_activity,
        cancel_activity,
        controller,
        notification_fixture,
        user_factory,
        teams_event_fixture,
        delete_notification,
    ):
        global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_activity"] = cancel_activity
        global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_external_activity"] = (
            cancel_external_creator_activity
        )

        data = self._get_user_and_activities(notification_fixture, user_factory, teams_event_fixture)
        other1_activity, other2_activity, activity = data["activities"]
        user1, user2 = data["users"]
        permission = Permission.objects.get_or_create(
            content_type=ContentType.objects.get_for_model(User), codename="is_internal_user"
        )[0]
        user2.user_permissions.add(permission)
        user_registry.reset_cache()
        user2_dto = UserDTO(metadata=user2.metadata, id=None)
        Activity.objects.filter(id=activity.id).update(creator=user1.profile)
        activity.refresh_from_db()
        activity_dto = activity._build_dto()
        activity_dto.delete_notification = delete_notification

        assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}

        # user 1 is an external user and it's the creator of the activity without an active subscription
        # user 2 is a internal participant
        controller.delete_activity(activity_dto, user2_dto)
        if delete_notification and cancel_external_creator_activity and not cancel_activity:
            assert set(Activity.objects.all()) == {other1_activity, other2_activity}
        else:
            assert set(Activity.objects.all()) == {other1_activity, other2_activity, activity}



from datetime import datetime, timedelta

import pytz
from dateutil import rrule
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.utils.rrules import convert_rrulestr_to_dict

from wbcrm.preferences import (
    get_maximum_allowed_recurrent_date,
    get_recurrence_maximum_count,
)


class Recurrence(CalendarItem):
    class ReoccuranceChoice(models.TextChoices):
        NEVER = "NEVER", _("Never")
        BUSINESS_DAILY = "RRULE:FREQ=DAILY;INTERVAL=1;WKST=MO;BYDAY=MO,TU,WE,TH,FR", _("Business Daily")
        DAILY = "RRULE:FREQ=DAILY", _("Daily")
        WEEKLY = "RRULE:FREQ=WEEKLY", _("Weekly")
        BIWEEKLY = "RRULE:FREQ=WEEKLY;INTERVAL=2", _("Bi-Weekly")
        MONTHLY = "RRULE:FREQ=MONTHLY", _("Monthly")
        QUARTERLY = "RRULE:FREQ=MONTHLY;INTERVAL=3", _("Quarterly")
        YEARLY = "RRULE:FREQ=YEARLY", _("Annually")

    parent_occurrence = models.ForeignKey(
        to="self",
        related_name="child_activities",
        null=True,
        blank=True,
        verbose_name=_("Parent Activity"),
        on_delete=models.deletion.DO_NOTHING,
    )
    propagate_for_all_children = models.BooleanField(
        default=False,
        verbose_name=_("Propagate for all following activities?"),
        help_text=_("If this is checked, changes will be propagated to the following activities."),
    )
    exclude_from_propagation = models.BooleanField(
        default=False,
        verbose_name=_("Exclude occurrence from propagation?"),
        help_text=_("If this is checked, changes will not be propagated on this activity."),
    )
    recurrence_end = models.DateField(
        verbose_name=_("Date"),
        null=True,
        blank=True,
        help_text=_(
            "Specifies until when an event is to be repeated. Is mutually exclusive with the Recurrence Count."
        ),
    )
    recurrence_count = models.IntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(365)],
        null=True,
        blank=True,
        verbose_name=_("Count"),
        help_text=_(
            "Specifies how often an activity should be repeated excluding the original activity. Is mutually exclusive with the end date. Limited to a maximum of 365 recurrences."
        ),
    )
    repeat_choice = models.CharField(
        max_length=56,
        default=ReoccuranceChoice.NEVER,
        choices=ReoccuranceChoice.choices,
        verbose_name=_("Recurrence Frequency"),
        help_text=_("Repeat activity at the specified frequency"),
    )

    class Meta:
        abstract = True

    @property
    def is_recurrent(self):
        return self.repeat_choice != Recurrence.ReoccuranceChoice.NEVER

    @property
    def is_root(self):
        return self.is_recurrent and not self.parent_occurrence

    @property
    def is_leaf(self):
        return self.is_recurrent and not self.next_occurrence

    @property
    def next_occurrence(self):
        if self.is_recurrent:
            parent_occurrence = self if self.is_root else self.parent_occurrence
            if qs := parent_occurrence.child_activities.filter(
                is_active=True, period__startswith__gt=self.period.lower
            ):
                return qs.earliest("period__startswith")

    @property
    def previous_occurrence(self):
        if self.is_recurrent and not self.is_root:
            if qs := self.parent_occurrence.child_activities.filter(
                is_active=True, period__startswith__lt=self.period.lower
            ):
                return qs.latest("period__startswith")
            else:
                return self.parent_occurrence
        # Else, it is a pivot and therefore, no previous occurrence should exist

    def save(self, *args, **kwargs):
        if not self.recurrence_end:
            self.recurrence_end = get_maximum_allowed_recurrent_date()
        if self.does_recurrence_need_cancellation():
            self.cancel_recurrence()
        super().save(*args, **kwargs)

    def delete(self, **kwargs):
        if self.propagate_for_all_children:
            self.forward_deletion()
        elif self.is_root and (next_occurrence := self.next_occurrence):
            next_occurrence.claim_parent_hood()
        super().delete(**kwargs)

    def _get_occurrence_start_datetimes(self, include_self: bool = False) -> list:
        """
        Returns a list with datetime values based on the recurrence options of an activity.

        :return:    list with datetime values
        """
        if self.is_recurrent:
            max_allowed_date = get_maximum_allowed_recurrent_date()
            max_allowed_count = get_recurrence_maximum_count()
            occurrence_count = min(
                self.recurrence_count + 1 if self.recurrence_count else max_allowed_count, max_allowed_count
            )
            end_date = min(
                self.recurrence_end + timedelta(days=1) if self.recurrence_end else max_allowed_date, max_allowed_date
            )
            end_date_time = datetime(end_date.year, end_date.month, end_date.day).astimezone(pytz.utc)
            rule_dict = convert_rrulestr_to_dict(
                self.repeat_choice, dtstart=self.period.lower, count=occurrence_count, until=end_date_time
            )

            start_datetimes = list(rrule.rrule(**rule_dict))
            if not include_self:
                start_date = self.period.lower.replace(microsecond=0)
                if self.period.lower in start_datetimes:
                    start_datetimes.remove(self.period.lower)
                elif start_date in start_datetimes:
                    start_datetimes.remove(start_date)
            return start_datetimes

    def _create_recurrence_child(self, *args):
        """
        Return a new child object based on self (parent/root)
        """
        raise NotImplementedError()

    def get_recurrent_valid_children(self):
        """
        Return a valid queryset representing the recurrent children
        """
        return self.child_activities.filter(exclude_from_propagation=False).order_by("period__startswith")

    def get_recurrent_invalid_children(self):
        valid = self.get_recurrent_valid_children().values_list("id", flat=True)
        return self.child_activities.all().exclude(id__in=valid).order_by("period__startswith")

    def _handle_recurrence_m2m_forwarding(self, child):
        """
        Call this when m2m data needs to be forward from the parent (self) to the passed child

        Args:
            child: Child to get the m2m data from
        """
        pass

    def does_recurrence_need_cancellation(self) -> bool:
        return False

    def cancel_recurrence(self):
        """
        keep the link to the root but exclude the occurrence from the list of valid children's to maintain, propagations will no longer be applied to this activity

        its a transition method, save needs to be called
        """
        self.exclude_from_propagation = True

    def claim_parent_hood(self):
        if self.is_recurrent and not self.is_root:
            new_batch = self._meta.model.objects.filter(
                Q(parent_occurrence=self.parent_occurrence) & Q(period__startswith__gt=self.period.lower)
            ).exclude(id=self.id)
            if new_batch.exists():
                new_batch.update(parent_occurrence=self)
            self._meta.model.objects.filter(id=self.id).update(parent_occurrence=None)

    def generate_occurrences(self, allow_reclaiming_root: bool = True):
        """
        Creation of the occurrences
        Existing occurrences whose start date is part of the list of occurrences dates obtained from the recurrence pattern are excluded
        Those not in the list are deleted
        """
        if self.is_recurrent:
            if not self.is_root and allow_reclaiming_root:
                self.claim_parent_hood()
                self.refresh_from_db()
            if self.is_root:
                if old_occurrences_dict := {occ.period.lower: occ.id for occ in self.child_activities.all()}:
                    occurrence_dates = set(self._get_occurrence_start_datetimes())
                    old_occurrences_dates = set(old_occurrences_dict.keys())
                    new_occurrence_dates = occurrence_dates.difference(old_occurrences_dates)
                    if diff_inv := old_occurrences_dates.difference(occurrence_dates):
                        self.forward_deletion(child_ids=[old_occurrences_dict[x] for x in diff_inv])
                else:
                    new_occurrence_dates = self._get_occurrence_start_datetimes()
                for start_datetime in new_occurrence_dates:
                    child = self._create_recurrence_child(start_datetime)
                    self._handle_recurrence_m2m_forwarding(child)
        return self.get_recurrent_valid_children()

    def forward_change(
        self,
        allow_reclaiming_root: bool = True,
        fields_to_forward: tuple[str, ...] = (
            "online_meeting",
            "visibility",
            "conference_room",
            "title",
            "description",
            "type",
            "importance",
            "reminder_choice",
            "location",
            "location_longitude",
            "location_latitude",
            "assigned_to",
        ),
        period_time_changed: bool = False,
    ):
        """
        Forward the changes to the following occurrences
        param: fields_to_forward: allows you to specify the fields to update
               allow_reclaiming_root: if True we split the occurrences and this occurrence becomes the parent of the following occurrences
               period_time_changed: boolean field to know if we need to update the period time or not
        """
        if self.is_recurrent and self.propagate_for_all_children:
            if not self.is_root and allow_reclaiming_root:
                self.claim_parent_hood()
                self.refresh_from_db()
            if self.is_root:
                for child in self.get_recurrent_valid_children().order_by("period__startswith"):
                    for field in fields_to_forward:
                        setattr(child, field, getattr(self, field))
                    if period_time_changed:
                        tz = child.period.lower.tzinfo
                        start_datetime = datetime.combine(
                            child.period.lower.date(), self.period.lower.time()
                        ).astimezone(tz)
                        child.period = TimestamptzRange(start_datetime, (start_datetime + self.duration))
                    child.propagate_for_all_children = False
                    child.save()
                    self._handle_recurrence_m2m_forwarding(child)
            self._meta.model.objects.filter(id=self.id).update(propagate_for_all_children=False)

    def forward_deletion(self, child_ids: tuple["str", ...] = ()):
        if self.is_recurrent:
            if self.is_root:
                occurrences = self.get_recurrent_valid_children()
            else:
                occurrences = self.parent_occurrence.get_recurrent_valid_children().filter(
                    period__startswith__gt=self.period.lower
                )
            if child_ids:
                occurrences = occurrences.filter(id__in=child_ids)
            if self.propagate_for_all_children:
                # We don't call delete but update to is_active=False in order to silently delete the child activities without trggering the signals
                occurrences.update(is_active=False)
                self._meta.model.objects.filter(id=self.id).update(propagate_for_all_children=False)
            else:
                for occurrence in occurrences:
                    occurrence.delete()



from django.apps import AppConfig
from django.db.models.signals import post_migrate


class WbSyncConfig(AppConfig):
    name = "wbcrm.synchronization"

    def ready(self) -> None:
        from wbcrm.synchronization.management import initialize_task

        post_migrate.connect(
            initialize_task,
            dispatch_uid="wbcrm.synchronization.initialize_task",
        )



import calendar
import datetime as dt

week_ahead = dt.date.today() + dt.timedelta(days=7)
week_before = dt.date.today() - dt.timedelta(days=7)
week_ahead_timestamp = calendar.timegm(week_ahead.timetuple()) * 1000
week_before_timestamp = calendar.timegm(week_before.timetuple()) * 1000
credentials = '{"url": "https://fake_url.io", "type": "service_account", "project_id": "fake_project_id", "private_key_id": "fake_private_key_id", "private_key": "fake_private_key", "client_email": "client_mail@serviceaccount.com", "client_id": "fake_client_id", "auth_uri": "https://accounts.google.com/o/oauth2/auth", "token_uri": "https://oauth2.googleapis.com/token", "auth_provider_x509_cert_url": "https://fake_auth_url", "client_x509_cert_url": "https://fake_cert_url"}'

person_metadata = {"google_backend": {"watch": {"expiration": str(week_ahead_timestamp)}}}
person_metadata_expired = {"google_backend": {"watch": {"expiration": str(week_before_timestamp)}, "expired": True}}

event = {
    "attendees": [
        {"displayName": "Foo", "email": "Foo@Foo.com", "responseStatus": "accepted"},
        {"displayName": "Bar", "email": "Bar@Bar.com", "responseStatus": "declined"},
        {"displayName": "Foo Bar", "email": "Foo@Bar.com", "responseStatus": "tentative"},
        {"email": "Bar@Foo.com", "responseStatus": "tentative"},
    ]
}

event_data = {
    "id": "test",
    "items": [],
    "start": {"date": "2022-12-06", "dateTime": "2022-12-06T17:25:00+0200", "timeZone": "UTC"},
    "end": {"date": "2022-12-06", "dateTime": "2022-12-06T18:25:00+0200", "timeZone": "UTC"},
}
event_list = [
    {"id": "1", "metaTest": "Parent", "originalStartTime": {"dateTime": "Fake Date Time"}},
    {"id": "2", "metaTest": "Child A", "originalStartTime": {"dateTime": "Fake Date Time A"}},
    {"id": "3", "metaTest": "Child B", "originalStartTime": {"dateTime": "Fake Date Time B"}},
    {"id": "4", "metaTest": "Child C", "originalStartTime": {"dateTime": "Fake Date Time C"}},
]


class EventService:
    def insert(self, calendarId, body):
        return ExecuteService(calendarId, body)

    def instances(self, calendarId, eventId):
        return ExecuteService(calendarId, eventId)

    def delete(self, calendarId, eventId):
        return ExecuteService(calendarId, eventId)

    def update(self, calendarId, eventId, body=event):
        return ExecuteService(calendarId, eventId)

    def patch(self, calendarId, eventId, body=event):
        return ExecuteService(calendarId, eventId)

    def get(self, calendarId, eventId, body=event):
        return ExecuteService(calendarId, eventId)

    def list(self, calendarId, pageToken, syncToken):
        return ExecuteService(calendarId, pageToken)

    def watch(self, calendarId, body=event):
        return ExecuteService(calendarId, event)


class ChannelsService:
    def stop(self, body):
        return ExecuteService("", body)


class ExecuteService:
    def __init__(self, calendarId, body):
        self.calendarId = calendarId
        self.body = body

    def execute(self):
        return event_data


class ServiceData:
    def events(self):
        return EventService

    def channels(self):
        return ChannelsService



# Generated by Django 5.0.9 on 2024-12-05 09:15

import django.db.models.deletion
from django.db import migrations, models
from tqdm import tqdm
from wbcore.contrib.directory.preferences import get_main_company


def clean_main_company(apps, schema_editor):
    Activity = apps.get_model("wbcrm", "Activity")
    main_company = get_main_company()
    Activity.companies.through.objects.filter(company_id=main_company.id).delete()


def compute_str(company):
    rep = company.computed_str
    if company.customer_status:
        rep += f" ({company.customer_status.title})"
    return rep


def create_through_relations(apps, schema_editor):
    Activity = apps.get_model("wbcrm", "Activity")
    ActivityCompanyThroughModel = apps.get_model("wbcrm", "ActivityCompanyThroughModel")
    qs = Activity.objects.all()
    for a in tqdm(qs, total=qs.count()):
        for c in a.companies.all():
            ActivityCompanyThroughModel.objects.create(
                activity=a, company=c, customer_status=c.customer_status, computed_str=compute_str(c)
            )


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0009_remove_entry_external_identfier_and_more"),
        ("wbcrm", "0015_alter_activity_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="companies",
            field=models.ManyToManyField(
                blank=True,
                help_text="The list of companies other than the main company that participate in this activity",
                related_name="company_participates",
                to="directory.company",
                verbose_name="Participating Companies",
            ),
        ),
        migrations.RunPython(clean_main_company),
        migrations.CreateModel(
            name="ActivityCompanyThroughModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "activity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="activity_companies",
                        to="wbcrm.activity",
                        verbose_name="Activity",
                    ),
                ),
                (
                    "company",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="activity_companies",
                        to="directory.company",
                        verbose_name="Company",
                    ),
                ),
                (
                    "customer_status",
                    models.ForeignKey(
                        blank=True,
                        help_text="The Customer Status at activity creation time",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="activity_companies",
                        to="directory.customerstatus",
                        verbose_name="Initial Customer Status",
                    ),
                ),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
            ],
        ),
        migrations.RunPython(create_through_relations),
        migrations.RemoveField(
            model_name="activity",
            name="companies",
        ),
        migrations.AddField(
            model_name="activity",
            name="companies",
            field=models.ManyToManyField(
                blank=True,
                help_text="The list of companies other than the main company that participate in this activity",
                related_name="company_participates",
                through="wbcrm.ActivityCompanyThroughModel",
                to="directory.company",
                verbose_name="Participating Companies",
            ),
        ),
    ]



import enum
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

from psycopg.types.range import TimestamptzRange
from wbcore.contrib.agenda.typings import ConferenceRoom
from wbcore.contrib.directory.typings import Person


@dataclass
class User:
    metadata: dict[str, Any]
    id: str = None

    def __eq__(self, other):
        if other:
            return self.id == other.id
        return super().__eq__(other)


@dataclass
class ParticipantStatus:
    class ParticipationStatus(enum.Enum):
        CANCELLED = "CANCELLED"
        MAYBE = "MAYBE"
        ATTENDS = "ATTENDS"
        NOTRESPONDED = "NOTRESPONDED"
        ATTENDS_DIGITALLY = "ATTENDS_DIGITALLY"
        PENDING_INVITATION = "PENDING_INVITATION"

    person: Person
    status_changed: datetime = None
    status: str = ParticipationStatus.PENDING_INVITATION.name
    activity: "Activity" = None
    id: str = None

    def __eq__(self, other):
        if other and (
            (self.id and other.id and self.id == other.id)
            or (self.person == other.person and self.activity == other.activity)
        ):
            return True
        return super().__eq__(other)


@dataclass
class Activity:
    class ReoccuranceChoice(enum.Enum):
        NEVER = "NEVER"
        BUSINESS_DAILY = "RRULE:FREQ=DAILY;INTERVAL=1;WKST=MO;BYDAY=MO,TU,WE,TH,FR"
        DAILY = "RRULE:FREQ=DAILY"
        WEEKLY = "RRULE:FREQ=WEEKLY"
        BIWEEKLY = "RRULE:FREQ=WEEKLY;INTERVAL=2"
        MONTHLY = "RRULE:FREQ=MONTHLY"
        QUARTERLY = "RRULE:FREQ=MONTHLY;INTERVAL=3"
        YEARLY = "RRULE:FREQ=YEARLY"

    class Visibility(enum.Enum):
        PUBLIC = "PUBLIC"
        PRIVATE = "PRIVATE"
        CONFIDENTIAL = "CONFIDENTIAL"

    class ReminderChoice(enum.Enum):
        NEVER = "NEVER"
        EVENT_TIME = "EVENT_TIME"
        MINUTES_5 = "MINUTES_5"
        MINUTES_15 = "MINUTES_15"
        MINUTES_30 = "MINUTES_30"
        HOURS_1 = "HOURS_1"
        HOURS_2 = "HOURS_2"
        HOURS_12 = "HOURS_12"
        WEEKS_1 = "WEEKS_1"

    @property
    def is_recurrent(self):
        return self.repeat_choice != self.ReoccuranceChoice.NEVER.name

    metadata: dict[str, Any]
    title: str
    period: TimestamptzRange = None
    description: str = ""
    participants: list["ParticipantStatus"] = field(default_factory=list)
    creator: Person = None
    visibility: str = Visibility.PUBLIC
    reminder_choice: str = ReminderChoice.MINUTES_15.name
    is_cancelled: bool = False
    all_day: bool = False
    online_meeting: bool = False
    location: str = None
    conference_room: ConferenceRoom = None
    id: str = None

    # parent_occurrence: "Activity" = None
    recurring_activities: list["Activity"] = field(default_factory=list)
    invalid_recurring_activities: list["Activity"] = field(default_factory=list)
    is_root: bool = False
    is_leaf: bool = False
    exclude_from_propagation: bool = False
    propagate_for_all_children: bool = False
    recurrence_end: datetime = None
    recurrence_count: int = 0
    repeat_choice: str = ReoccuranceChoice.NEVER.name
    delete_notification: bool = False

    def __eq__(self, other):
        if other and (self.id and other.id and self.id == other.id):
            return True
        return super().__eq__(other)



# Generated by Django 4.1.8 on 2023-04-21 08:27

import django_fsm
from django.db import migrations, models


def move_external_id_in_metadata(apps, schema):
    Activity = apps.get_model("wbcrm", "Activity")
    for activity in Activity.objects.filter(external_id__isnull=False):
        if activity.parent_activity:
            metadata = {"outlook": {"event_id": activity.external_id}}
        else:
            metadata = {"outlook": {"event_uid": activity.external_id}}
        activity.metadata = metadata
        activity.save()


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0002_alter_activity_repeat_choice"),
    ]

    operations = [
        migrations.AddField(
            model_name="activity",
            name="exclude_from_propagation",
            field=models.BooleanField(
                default=False,
                help_text="If this is checked, changes will not be propagated on this activity.",
                verbose_name="Exclude occurrence from propagation?",
            ),
        ),
        migrations.AddField(
            model_name="activity",
            name="metadata",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.RemoveField(
            model_name="activity",
            name="external_id",
        ),
        migrations.AlterField(
            model_name="activityparticipant",
            name="participation_status",
            field=django_fsm.FSMField(
                choices=[
                    ("CANCELLED", "Cancelled"),
                    ("MAYBE", "Maybe"),
                    ("ATTENDS", "Attends"),
                    ("NOTRESPONDED", "Not Responded"),
                    ("ATTENDS_DIGITALLY", "Attends Digitally"),
                ],
                default="NOTRESPONDED",
                max_length=50,
                verbose_name="Participation Status",
            ),
        ),
        migrations.RenameField(
            model_name="activity",
            old_name="parent_activity",
            new_name="parent_occurrence",
        ),
    ]



# from wbcore.contrib.currency.tests import factories

from .accounts import AccountFactory, AccountRoleFactory, AccountRoleTypeFactory, AccountWithOwnerFactory
from .activities import (
    ActivityCompanyFactory,
    ActivityFactory,
    ActivityParticipantFactory,
    ActivityPersonFactory,
    ActivityTypeCALLFactory,
    ActivityTypeFactory,
    RecurringActivityFactory,
)
from .groups import GroupFactory
from .products import ProductFactory



from selenium.webdriver.remote.webdriver import WebDriver
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.directory.models import Person
from wbcore.test import (
    click_element_by_path,
    fill_out_form_fields,
    open_create_instance,
)

from wbcrm.factories import ActivityFactory, ActivityTypeFactory
from wbcrm.models import Activity, ActivityType
from wbcrm.serializers import ActivityModelSerializer


def set_up_activities(creator: Person = None):
    """Sets up three activities for testing purposes with fixed participants, types and titles.

    Args:
        creator (Person | None, optional): The person that is used as the activities creator. Defaults to None.
    """
    type_a = ActivityTypeFactory(title="Meeting")
    type_b = ActivityTypeFactory(title="Call")
    participant_a = PersonFactory(last_name="Turing")
    participant_b = PersonFactory(last_name="Zuse")
    participant_c = PersonFactory(last_name="Lovelace")
    activity_creator = creator if creator else PersonFactory()
    ActivityFactory(
        title="Activity A",
        type=type_a,
        creator=activity_creator,
        preceded_by=None,
        participants=[participant_a, participant_c],
        status=Activity.Status.PLANNED,
    )
    ActivityFactory(
        title="Activity B",
        type=type_a,
        creator=activity_creator,
        preceded_by=None,
        participants=[participant_a, participant_b],
        status=Activity.Status.FINISHED,
    )
    ActivityFactory(
        title="Activity C",
        type=type_b,
        creator=activity_creator,
        preceded_by=None,
        participants=[participant_b, participant_c],
        status=Activity.Status.PLANNED,
    )


def create_new_activity_instance(
    driver: WebDriver,
    field_list: list[str],
    type_title="",
    activity_title="",
    is_create_instance_open=True,
) -> Activity:
    """A function that automatically creates a new activity for selenium e2e-tests.
    After creating the instance this function will close the create-widget.

    Args:
        driver (WebDriver): The Selenium webdriver.
        field_list (list[str]): List of fields to be filled in the creation mask. The field names must match the names in the ActivitySerializer.
        type_title (str, optional): The title of the activity type. Defaults to "".
        activity_title (str, optional): The title of the activity. Defaults to "".
        is_create_instance_open (bool, optional): Should be true if the create-widget is already open. Defaults to True.
    """
    if not is_create_instance_open:
        open_create_instance(driver, "CRM", "Create Activity")

    type_title = type_title if type_title else "Test Activity Type"
    activity_title = activity_title if activity_title else "Test Activity"
    if ActivityType.objects.filter(title=activity_title).exists():
        activity_type = ActivityType.objects.get(title=activity_title)
    else:
        activity_type = ActivityTypeFactory(title=type_title)
    activity = ActivityFactory.build(title=activity_title, type=activity_type)
    serializer = ActivityModelSerializer(activity)
    fill_out_form_fields(driver, serializer, field_list, activity)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return activity



from datetime import timedelta

import pytest
from django.apps import apps
from django.utils import timezone
from django_fsm import TransitionNotAllowed
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.authentication.factories import (
    AuthenticatedPersonFactory,
    UserFactory,
)
from wbcore.contrib.directory.factories import PersonWithEmployerFactory
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.icons import WBIcon

from wbcrm.models import Activity, add_employer_to_activities


@pytest.mark.django_db
class TestSpecificModelsActivities:
    def test_get_color_map(self, activity_factory):
        activity1 = activity_factory()
        assert activity1.Status.get_color_map()

    def test_get_minutes_correspondance(self, activity_factory):
        activity1 = activity_factory()
        assert activity1.ReminderChoice.get_minutes_correspondance("NEVER")
        assert activity1.ReminderChoice.get_minutes_correspondance("EVENT_TIME") == 0
        assert activity1.ReminderChoice.get_minutes_correspondance("MINUTES_5")
        assert activity1.ReminderChoice.get_minutes_correspondance("MINUTES_15")
        assert activity1.ReminderChoice.get_minutes_correspondance("MINUTES_30")
        assert activity1.ReminderChoice.get_minutes_correspondance("HOURS_1")
        assert activity1.ReminderChoice.get_minutes_correspondance("HOURS_2")
        assert activity1.ReminderChoice.get_minutes_correspondance("HOURS_12")
        assert activity1.ReminderChoice.get_minutes_correspondance("WEEKS_1")

    def test_str(self, activity_factory):
        assert activity_factory(preceded_by=None).__str__()

    def test_get_participants(self, activity_factory, person_factory, internal_user_factory, company_factory):
        person1 = person_factory()
        company1 = company_factory()
        if apps.is_installed("wbhuman_resources"):
            from wbhuman_resources.factories import EmployeeHumanResourceFactory

            person2 = AuthenticatedPersonFactory(employers=(company1,))
            EmployeeHumanResourceFactory(profile=person2)
        else:
            person2 = person_factory(employers=(company1,))
            u = internal_user_factory.create()
            u.profile = person2
            u.save()
        activity1 = activity_factory(participants=(person1,))
        assert activity1.get_participants().count() == 1
        assert activity1.get_participants()[0] == person1

    def test_get_activities_for_user(self, activity_factory):
        activity0 = activity_factory(preceded_by=None)
        user = UserFactory(is_active=True, is_superuser=True)
        qs = activity0.get_activities_for_user(user)
        assert qs.count() == 1

        user2 = UserFactory(is_active=True)
        qs2 = activity0.get_activities_for_user(user2)
        assert qs2.count() == 0

    def test_add_group_to_activity(self, activity_factory, group_factory, person_factory, company_factory):
        activity_participant_a = person_factory()
        group_member_a = person_factory()
        member_a_entry = Entry.objects.get(pk=group_member_a.pk)
        group_member_b = company_factory()
        member_b_entry = Entry.objects.get(pk=group_member_b.pk)
        group_a = group_factory(members=(member_a_entry, member_b_entry))

        activity = activity_factory(
            participants=[
                activity_participant_a,
            ]
        )
        assert activity.participants.filter(pk=activity_participant_a.pk).exists()
        assert not activity.participants.filter(pk=group_member_a.pk).exists()
        assert not activity.companies.filter(pk=group_member_b.pk).exists()

        activity.groups.set((group_a,))
        activity.save()

        assert activity.participants.filter(pk=activity_participant_a.pk).exists()
        assert activity.participants.filter(pk=group_member_a.pk).exists()
        assert activity.companies.filter(pk=group_member_b.pk).exists()

    def test_remove_group_to_activity(self, activity_factory, group_factory, person_factory, company_factory):
        group_member_a = person_factory()
        group_member_b = person_factory()
        group_member_c = person_factory()
        group_member_d = company_factory()
        group_a = group_factory(members=(group_member_a.entry_ptr, group_member_c.entry_ptr, group_member_d.entry_ptr))
        group_b = group_factory(members=(group_member_b.entry_ptr, group_member_c.entry_ptr))

        activity = activity_factory(groups=[group_a, group_b])

        assert activity.participants.filter(pk=group_member_a.pk).exists()
        assert activity.participants.filter(pk=group_member_b.pk).exists()
        assert activity.participants.filter(pk=group_member_c.pk).exists()
        assert activity.companies.filter(pk=group_member_d.pk).exists()

        activity.groups.remove(group_a.pk)
        activity.save()

        assert not activity.participants.filter(pk=group_member_a.pk).exists()
        assert activity.participants.filter(pk=group_member_b.pk).exists()
        assert activity.participants.filter(pk=group_member_c.pk).exists()
        assert not activity.companies.filter(pk=group_member_d.pk).exists()

    def test_add_participants_employer(self, activity_factory, person_factory):
        person_a = PersonWithEmployerFactory()
        person_b = person_factory()
        employer_a = person_a.employer.get(primary=True).employer
        activity_a = activity_factory(participants=(person_a, person_b))
        assert person_a in activity_a.participants.all()
        assert person_b in activity_a.participants.all()
        assert employer_a in activity_a.companies.all()

    def test_add_participants_employer_on_update(self, activity_factory, person_factory):
        person_a = PersonWithEmployerFactory()
        person_b = person_factory()
        employer_a = person_a.employer.get(primary=True).employer
        activity_a = activity_factory(participants=(person_b,))
        assert person_a not in activity_a.participants.all()
        assert person_b in activity_a.participants.all()
        assert employer_a not in activity_a.companies.all()
        activity_a.participants.set([person_a, person_b])
        assert person_a in activity_a.participants.all()
        assert person_b in activity_a.participants.all()
        assert employer_a in activity_a.companies.all()

    def test_add_participants_employer_to_canceled_activity(self, activity_factory, person_factory):
        person_a = PersonWithEmployerFactory()
        person_b = person_factory()
        activity_a = activity_factory(participants=(person_a, person_b), status=Activity.Status.CANCELLED)
        assert set(activity_a.participants.all()) == {person_a, person_b}

    def test_participant_cannot_attend_digitally(
        self, activity_participant_factory, activity_factory, conference_room_factory
    ):
        room = conference_room_factory(is_videoconference_capable=False)
        activity = activity_factory(conference_room=room)
        obj = activity_participant_factory(activity=activity)
        with pytest.raises(TransitionNotAllowed):
            obj.attendsdigitally()

    def test_participant_can_attend_digitally(self, activity_participant_factory, activity_factory):
        activity = activity_factory()
        obj = activity_participant_factory(activity=activity)
        obj.attendsdigitally()
        assert True

    def test_get_casted_calendar_item(self, activity_factory):
        activity = activity_factory()
        calendar_item = CalendarItem.objects.get(id=activity.id)
        assert calendar_item.get_casted_calendar_item() == activity

    def test_add_employer_to_activities(
        self, activity_factory, employer_employee_relationship_factory, person_factory
    ):
        employee = person_factory()
        act1 = activity_factory(
            status=Activity.Status.PLANNED, start=timezone.now() + timedelta(1), participants=[employee]
        )
        act2 = activity_factory(
            status=Activity.Status.REVIEWED, start=timezone.now() + timedelta(1), participants=[employee]
        )
        act3 = activity_factory(
            status=Activity.Status.PLANNED, start=timezone.now() - timedelta(1), participants=[employee]
        )
        eer = employer_employee_relationship_factory(employee=employee)
        add_employer_to_activities(eer.id)
        act1.refresh_from_db()
        act2.refresh_from_db()
        act3.refresh_from_db()
        assert eer.employer in act1.companies.all()
        assert eer.employer not in act2.companies.all()
        assert eer.employer not in act3.companies.all()
        assert eer.employer.entry_ptr in act1.entities.all()
        assert eer.employer.entry_ptr not in act2.entities.all()
        assert eer.employer.entry_ptr not in act3.entities.all()

    def test_set_color_icons(self, activity_factory, calendar_item_factory):
        activity = activity_factory(type__icon=WBIcon.EURO.icon)
        assert activity.color == activity.type.color
        assert activity.icon == activity.type.icon

        activity_type = activity.type
        activity_type.icon = WBIcon.LOCATION.icon
        activity_type.color = "#42f444"
        activity_type.save()
        activity.refresh_from_db()
        assert activity.color == activity_type.color
        assert activity.icon == activity_type.icon

        calendar_item = calendar_item_factory()
        assert calendar_item.color
        assert calendar_item.icon

    def test_main_company_removed_from_m2m(self, activity, company_factory):
        external_company = company_factory.create()
        activity.companies.add(external_company)
        assert set(activity.companies.all()) == {
            external_company,
        }

        main_company = company_factory.create()
        global_preferences_registry.manager()["directory__main_company"] = main_company.id
        activity.companies.add(main_company)
        assert set(activity.companies.all()) == {
            external_company,
        }



from datetime import date, timedelta

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from faker import Faker
from psycopg.types.range import DateRange
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.directory.signals import deactivate_profile

from wbcrm.models import Account, AccountRole, AccountRoleValidity

fake = Faker()


@pytest.mark.django_db
class TestAccountModel:
    def test_init(self, account):
        assert account.id is not None

    def test_close_account_set_inactive(self, account):
        assert account.is_active
        account.status = Account.Status.CLOSE
        account.save()
        assert not account.is_active

    def test_disable_account_set_status_close(self, account):
        assert account.status == Account.Status.OPEN
        account.delete()
        assert not account.is_active
        assert account.status == Account.Status.CLOSE

    def test_account_with_owner_create_account_role(self, account_factory, entry_factory):
        customer = entry_factory.create()
        account = account_factory.create(owner=customer)
        assert account.roles.filter(entry=customer, role_type__key="customer").exists()

    def test_leaf_account_automatically_set_to_terminal(self, account_factory):
        parent_account = account_factory.create()

        assert parent_account.is_terminal_account is True
        child_account1 = account_factory.create(parent=parent_account, is_terminal_account=False)
        child_account2 = account_factory.create(parent=parent_account, is_terminal_account=False)

        parent_account.refresh_from_db()
        child_account1.refresh_from_db()
        child_account2.refresh_from_db()

        assert parent_account.is_terminal_account is False
        assert child_account1.is_terminal_account is True
        assert child_account2.is_terminal_account is True

        child_account1.delete(no_deletion=False)
        parent_account.refresh_from_db()
        assert parent_account.is_terminal_account is False  # because there is still another child

        child_account2.delete(no_deletion=False)
        parent_account.refresh_from_db()
        assert parent_account.is_terminal_account is True  # There is no children left so we expect it to become leaf

    def test_disabled_parent_account_disabled_children(self, account_factory):
        parent_account = account_factory.create()
        child_account = account_factory.create(parent=parent_account)
        assert child_account.is_active
        assert parent_account.is_active

        parent_account.is_active = False
        parent_account.save()

        child_account.refresh_from_db()
        assert not child_account.is_active

    def test_private_root_account_set_children_as_private(self, account_factory):
        parent_account = account_factory.create(is_public=True)
        child_account = account_factory.create(parent=parent_account, is_public=True)
        parent_account.is_public = False
        parent_account.save()
        child_account.refresh_from_db()
        assert not child_account.is_public

    def test_can_manage(self, account, user):
        assert not account.can_administrate(user)

        perm = Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
        user.user_permissions.add(perm)

        user = get_user_model().objects.get(
            id=user.id
        )  # we refetch user to clear perm cache, doesn't happen with refresh_from_db
        assert account.can_administrate(user)

    def test_get_inherited_roles_for_account(self, account_factory, account_role_factory):
        parent_account = account_factory.create()
        parent_role = account_role_factory.create(account=parent_account)
        child_account = account_factory.create(parent=parent_account)
        child_role = account_role_factory.create(account=child_account)
        assert set(child_account.get_inherited_roles_for_account()) == {parent_role}
        assert set(child_account.get_inherited_roles_for_account(include_self=True)) == {child_role, parent_role}

    @pytest.mark.parametrize("validity_date,account__is_public", [(fake.date_object(), False)])
    def test_can_see_account(self, user, account, account_role_factory, validity_date):
        # Check user doesn't see permission on the account by default
        assert not account.can_see_account(user, validity_date)
        role = account_role_factory.create(account=account, entry=Entry.objects.get(id=user.profile.id))
        # with a valid role, they can see it
        assert account.can_see_account(user, validity_date=validity_date)

        # with an unvalid role, they cannot see it
        role.validity_set.first().timespan = DateRange(date.min, validity_date - timedelta(days=1))
        role.validity_set.first().save()
        assert account.can_see_account(user, validity_date=validity_date)

    @pytest.mark.parametrize("account__is_public", [True, False])
    def test_can_see_account_with_internal_employee(self, internal_user_factory, account):
        # Check user doesn't see permission on the account by default

        # with being an internal user, they can see it
        assert account.can_see_account(internal_user_factory.create()) == account.is_public

    @pytest.mark.parametrize("account__is_public", [True])
    def test_normaluser_cannot_see_account_public_account(self, user, account):
        assert not account.can_see_account(user)

    def test_get_accounts_for_customer(self, account_factory, company_factory, employer_employee_relationship_factory):
        owner1 = company_factory.create()
        employee1 = employer_employee_relationship_factory.create(employer=owner1).employee
        owner2 = company_factory.create()
        employee2 = employer_employee_relationship_factory.create(employer=owner2).employee

        parent_account = account_factory.create(owner=owner1)
        other_parent_account = account_factory.create(owner=owner1)
        child_account = account_factory.create(parent=parent_account, owner=owner2)
        # test if all child account also shows in the queryset as well as the direct related account for that owner
        assert set(Account.get_accounts_for_customer(owner1.entry_ptr)) == {
            parent_account,
            other_parent_account,
            child_account,
        }
        assert set(Account.get_accounts_for_customer(employee1.entry_ptr)) == {
            parent_account,
            other_parent_account,
            child_account,
        }
        assert set(Account.get_accounts_for_customer(owner2.entry_ptr)) == {child_account}
        assert set(Account.get_accounts_for_customer(employee2.entry_ptr)) == {child_account}

    def test_get_managed_accounts_for_entry(self, entry, account_factory, account_role_factory):
        main_role = account_role_factory.create(entry=entry)
        child_account = account_factory.create(parent=main_role.account)
        other_account = account_factory.create()  # noqa
        other_role = account_role_factory.create()  # noqa
        assert set(Account.get_managed_accounts_for_entry(entry)) == {main_role.account, child_account}

    def test_handle_user_deactivation(self, account_role_factory, person_factory):
        profile = person_factory.create()
        substitute_profile = person_factory.create()

        valid_role = account_role_factory.create(entry=Entry.objects.get(id=profile.id))
        unvalid_role = account_role_factory.create(
            entry=Entry.objects.get(id=profile.id), visibility_daterange=DateRange(date.min, date.today())
        )

        deactivate_profile.send(AccountRole, instance=profile, substitute_profile=substitute_profile)
        valid_role.refresh_from_db()
        unvalid_role.refresh_from_db()
        assert valid_role.validity_set.first().timespan == DateRange(date.min, date.today() - timedelta(days=1))
        assert valid_role.entry.id == profile.id
        assert unvalid_role.entry.id == profile.id

        new_role = AccountRole.objects.get(entry_id=substitute_profile.id, account=valid_role.account)
        assert new_role.validity_set.first().timespan == DateRange(date.today() - timedelta(days=1), date.max)

    @pytest.mark.parametrize(
        "validity_date,is_internal_user, is_superuser",
        [
            (fake.date_object(), True, True),
            (fake.date_object(), False, True),
            (fake.date_object(), True, False),
            (fake.date_object(), False, False),
        ],
    )
    def test_filter_for_user(
        self,
        user_factory,
        internal_user_factory,
        account_factory,
        account_role_factory,
        validity_date,
        is_internal_user,
        is_superuser,
    ):
        # We assume role logic are tested in the AccountRole manager unit test
        # so we test here only:
        # - chain filtering
        # - Return account where user has a valid role
        # - and Return public account if user is superuser or internal

        if is_internal_user:
            user = internal_user_factory.create()
        else:
            user = user_factory.create()
        if is_superuser:
            user.user_permissions.add(
                Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
            )
        profile_entry = Entry.objects.get(id=user.profile.id)
        public_account = account_factory.create(is_public=True)
        private_account = account_factory.create(is_public=False)  # noqa
        valid_role = account_role_factory.create(account__is_public=fake.pybool(), entry=profile_entry)
        children_valid_role_account = account_factory.create(
            parent=valid_role.account, is_public=True
        )  # check that entry with role on the parent account can access the public children account
        account_factory.create(
            parent=valid_role.account, is_public=False
        )  # check that entry with role on the parent account cannot access private children account

        unvalid_role = account_role_factory.create(  # noqa
            account__is_public=False,
            entry=profile_entry,
            visibility_daterange=DateRange(date.min, validity_date - timedelta(days=1)),
        )
        user = get_user_model().objects.get(id=user.id)
        if is_superuser:
            assert set(Account.objects.filter_for_user(user, validity_date)) == set(Account.objects.all())
        elif is_internal_user:
            assert set(Account.objects.filter_for_user(user, validity_date)) == {
                public_account,
                valid_role.account,
                children_valid_role_account,
            }
            assert set(Account.objects.filter(id=public_account.id).filter_for_user(user, validity_date)) == {
                public_account
            }  # basic test filter chainage
        else:
            assert set(Account.objects.filter_for_user(user, validity_date)) == {
                valid_role.account,
                children_valid_role_account,
            }

    def test_cannot_merge_account_with_children(self, account_factory):
        """
        Check that we cannot merge an account that still has children
        """
        base_account = account_factory.create()
        merged_account = account_factory.create()
        children_merged_account = account_factory.create(parent=merged_account)
        base_account.merge(merged_account)
        merged_account.refresh_from_db()
        assert merged_account

        # but can merge the children into the parent account
        merged_account.merge(children_merged_account)
        with pytest.raises(Account.DoesNotExist):
            children_merged_account.refresh_from_db()

    def test_cannot_merge_account_on_different_trees(self, account_factory):
        """
        Check that we cannot merge an account that still has children
        """
        base_account = account_factory.create(parent=account_factory.create())
        merged_account = account_factory.create(parent=account_factory.create())
        account_factory.create(parent=merged_account)
        base_account.merge(merged_account)
        merged_account.refresh_from_db()
        assert merged_account

    def test_account_merging(self, account_factory, account_role_factory, user):
        # TODO implemetns for commission
        pivot_date = date(2023, 1, 1)
        base_account = account_factory.create()
        base_account_role = account_role_factory.create(
            account=base_account, visibility_daterange=DateRange(date.min, pivot_date)
        )
        base_account_role_validity = base_account_role.validity_set.first()

        merged_account = account_factory.create()
        merged_account_role_overlapping = account_role_factory.create(
            account=merged_account, entry=base_account_role.entry, visibility_daterange=DateRange(pivot_date, date.max)
        )
        merged_account_role_overlapping.authorized_hidden_users.add(user)
        merged_account_role_validity = merged_account_role_overlapping.validity_set.first()

        merged_account_role_different = account_role_factory.create(account=merged_account)
        merged_account_role_different_validity = merged_account_role_different.validity_set.first()

        base_account.merge(merged_account)

        # test that the none overlapping role were forwarded to the base account
        assert set(base_account.roles.all()) == {base_account_role, merged_account_role_different}

        # test that the validity set for the base role got the new validity from the base merged account role
        assert set(base_account_role.validity_set.all()) == {base_account_role_validity, merged_account_role_validity}

        # test that the authorized user set on the merged account role were forwarded to the base account
        assert set(base_account_role.authorized_hidden_users.all()) == {user}

        # test that the merged account was deleted
        with pytest.raises(Account.DoesNotExist):
            merged_account.refresh_from_db()

        # test that the non overlapping role correctly follows the role
        role = AccountRole.objects.get(account=base_account, entry=merged_account_role_different.entry)
        assert set(role.validity_set.all()) == {merged_account_role_different_validity}


@pytest.mark.django_db
class TestAccountRoleModel:
    def test_init(self, account_role):
        assert account_role.id is not None
        validity = account_role.validity_set.first()
        assert validity
        assert validity.timespan == DateRange(date.min, date.max)

    def test_hidden_role_set_account_to_private(self, account_factory, account_role_factory):
        public_account = account_factory.create(is_public=True)
        role = account_role_factory.create(account=public_account, is_hidden=True)
        assert not role.account.is_public

    # def test_filter_for_user(self, ):
    #     # here we want to check that the chained filtering works and
    #     # - Internal or super user can see their roles + public account roles if there are unhidden + hidden role if they have the right to see them
    #     # - Otherwise, can only see direct roles
    #     # - Check if the annotation is_currently_valid is there
    @pytest.mark.parametrize(
        "is_superuser,is_internal_user",
        [
            (True, True),
            (True, False),
            (False, True),
            (False, False),
        ],
    )
    def test_filter_related_account_roles_for_internal_user(
        self, user_factory, internal_user_factory, account_role_factory, is_superuser, is_internal_user
    ):
        if is_internal_user:
            user = internal_user_factory.create()
        else:
            user = user_factory.create()
        if is_superuser:
            user.user_permissions.add(
                Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
            )
        role_unhidden_public_account = account_role_factory.create(is_hidden=False, account__is_public=True)
        role_unhidden_private_account = account_role_factory.create(is_hidden=False, account__is_public=False)  # noqa
        role_hidden_public_account_without_privilege = account_role_factory.create(  # noqa
            is_hidden=True, account__is_public=True
        )

        role_hidden_private_account_with_privilege = account_role_factory.create(
            is_hidden=True, account__is_public=False, authorized_hidden_users=[user]
        )
        role_direct = account_role_factory.create(
            entry=Entry.objects.get(id=user.profile.id),
            account__is_public=fake.pybool(),
        )
        role_indirect_private_account = account_role_factory.create(account__is_public=False)  # noqa

        if is_superuser:
            assert set(AccountRole.objects.filter_for_user(user)) == set(AccountRole.objects.all())
        elif is_internal_user:
            assert set(AccountRole.objects.filter_for_user(user)) == {
                role_unhidden_public_account,
                role_hidden_private_account_with_privilege,
                role_direct,
            }
            assert set(AccountRole.objects.filter(is_hidden=True).filter_for_user(user)) == {
                role_hidden_private_account_with_privilege
            }  # Test that chain filtering works
        else:
            assert set(AccountRole.objects.filter_for_user(user)) == {role_direct}


@pytest.mark.django_db
class TestAccountRoleValidityModel:
    @pytest.mark.parametrize("validity_date", [fake.date_object()])
    def test_get_role_validity_subquery(self, account_role, validity_date):
        assert AccountRole.objects.count() == 1
        assert (
            AccountRole.objects.annotate(is_valid=AccountRoleValidity.get_role_validity_subquery(validity_date))
            .filter(is_valid=True)
            .count()
            == 1
        )

        validity = account_role.validity_set.first()
        validity.timespan = DateRange(date.min, validity_date - timedelta(days=1))
        validity.save()
        assert (
            not AccountRole.objects.annotate(is_valid=AccountRoleValidity.get_role_validity_subquery(validity_date))
            .filter(is_valid=True)
            .exists()
        )



from django.forms import ValidationError
from django.utils.translation import gettext_lazy as _
from slugify import slugify
from wbcore import serializers
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import CompanyRepresentationSerializer

from wbcrm.models import Product, ProductCompanyRelationship


class ProductRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcrm:productrepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcrm:product-detail")

    class Meta:
        model = Product
        fields = (
            "id",
            "title",
            "computed_str",
            "_detail",
            "is_competitor",
        )


class ProductCompanyRelationshipModelSerializer(serializers.ModelSerializer):
    _product = ProductRepresentationSerializer(source="product")
    competitor_product = wb_serializers.BooleanField(read_only=True)

    class Meta:
        model = ProductCompanyRelationship
        fields = ("id", "product", "_product", "company", "competitor_product")


class ProductModelSerializer(serializers.ModelSerializer):
    _prospects = CompanyRepresentationSerializer(source="prospects", many=True)

    def validate(self, data):
        title = data.get("title")
        competitor = data.get("is_competitor")
        if title:
            product = Product.objects.filter(is_competitor=competitor, slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                product = product.exclude(id=obj.id)
            if product.exists():
                product_type = _("competitor ") if competitor else ""
                raise ValidationError({"title": _("Cannot add a duplicate {}product.").format(product_type)})
        return data

    class Meta:
        model = Product
        fields = (
            "id",
            "title",
            "prospects",
            "_prospects",
            "is_competitor",
        )



from typing import TYPE_CHECKING

from langchain_core.messages import HumanMessage, SystemMessage
from pydantic import BaseModel, Field
from wbcore.contrib.ai.llm.config import LLMConfig
from wbcrm.models.activities import Activity

if TYPE_CHECKING:
    from wbcrm.models import Account


class AccountRelationshipResponseModel(BaseModel):
    relationship_status: int = Field(
        ...,
        ge=1,
        le=5,
        description="Rate the customer relationship status from 1 to 5. 1 being the cold and 5 being the hot.",
    )
    relationship_summary: str = Field(
        ...,
        description="Briefly summarize the current state of the relationship and recent interactions. Also include any additional information that might be relevant.",
    )
    action_plan: str = Field(
        ..., description="Provide the next recommended actions or steps to engage with the customer."
    )


def _get_query(account: "Account"):
    details = []
    if account.owner:
        for activity in Activity.objects.filter(companies__id__in=[account.owner.id], status="REVIEWED"):
            details.append((activity.summary, activity.heat, activity.period))
    return {"title": account.title, "details": details}


analyze_relationship = LLMConfig["Account"](
    key="analyze_relationship",
    prompt=[
        SystemMessage(
            content="Based on the recent interactions/activities, current relationship health, and planned actions with the customer, provide a status score (1 to 5, where 1 is cold and 5 is hot), a summary of the relationship, and recommended next steps for maintaining or improving the relationship. Keep in mind that more information might become available later to refine these insights. Please include all information that is relevant to the relationship in the summary. Regarding the activities/interactions, older interactions are less relevant than recent ones."
        ),
        HumanMessage(
            content="The title of the account: {title} and the list of activities details (tuple of summary, sentiment heat and period): {details}",
        ),
    ],
    on_save=True,
    on_condition=lambda act: (act.status == act.Status.OPEN) and act.owner is not None and act.is_root_node(),
    output_model=AccountRelationshipResponseModel,
    query=_get_query,
)



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class ActivityTypeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Activity Types")

    def get_create_title(self):
        return _("New Activity Type")

    def get_instance_title(self):
        return _("Activity Type")


class ActivityTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        if "pk" in self.view.kwargs:
            if self.view.is_private_for_user:
                return _("Private Activity")

            elif self.view.is_confidential_for_user:
                return _("Confidential Activity")

        return _("Activity: {{title}}")

    def get_delete_title(self):
        if "pk" in self.view.kwargs:
            if self.view.is_private_for_user:
                return _("Delete Private Activity")

            elif self.view.is_confidential_for_user:
                return _("Delete Confidential Activity")

        return _("Delete Activity: {{title}}")

    def get_list_title(self):
        if self.view.participants.exists():
            if self.view.participants.count() == 1:
                return _("Activities for {person}").format(person=str(self.view.participants.first()))
            else:
                return _("Activities for Multiple Persons")

        if self.view.companies.exists():
            return _("Activities for {}").format(", ".join(self.view.companies.values_list("name", flat=True)))
        return _("Activities")

    def get_create_title(self):
        if self.view.entry:
            return _("New Activity for {entry}").format(entry=self.view.entry.computed_str)
        return _("New Activity")


class ActivityParticipantTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Participation Status")


class ActivityChartModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Activity Chart")



# Generated by Django 4.1.9 on 2023-07-05 16:48
import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
import django_fsm
import mptt.fields
import wbcore.utils.strings
from django.conf import settings
from django.contrib.postgres.operations import BtreeGistExtension
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("directory", "0004_entry_is_draft_entry"),
        ("wbcrm", "0004_alter_activity_status"),
    ]

    operations = [
        BtreeGistExtension(),
        migrations.CreateModel(
            name="Account",
            fields=[
                ("reference_id", models.PositiveIntegerField(unique=True)),
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("is_active", models.BooleanField(default=True)),
                ("deletion_datetime", models.DateTimeField(blank=True, null=True)),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[("PENDING", "Pending"), ("OPEN", "Open"), ("CLOSE", "Close")],
                        default="CLOSE",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                (
                    "is_terminal_account",
                    models.BooleanField(
                        default=False,
                        help_text="If true, sales or revenue can happen in this account",
                        verbose_name="Terminal Account",
                    ),
                ),
                (
                    "is_public",
                    models.BooleanField(
                        default=False,
                        help_text="If True, all internal users can access this account",
                        verbose_name="Public",
                    ),
                ),
                ("lft", models.PositiveIntegerField(editable=False)),
                ("rght", models.PositiveIntegerField(editable=False)),
                ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
                ("level", models.PositiveIntegerField(editable=False)),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="accounts",
                        to="directory.entry",
                        verbose_name="Owner",
                    ),
                ),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="wbcrm.account",
                        verbose_name="Parent Account",
                    ),
                ),
            ],
            options={
                "verbose_name": "Account",
                "verbose_name_plural": "Accounts",
                "permissions": [("administrate_account", "Administrate Account")],
            },
            bases=(wbcore.utils.strings.ReferenceIDMixin, models.Model),
        ),
        migrations.CreateModel(
            name="AccountRole",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("weighting", models.FloatField(default=1, verbose_name="Weight")),
                (
                    "is_hidden",
                    models.BooleanField(
                        default=False,
                        help_text="If True, this role is hidden and can be seen only by authorized people",
                        verbose_name="Hidden",
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "authorized_hidden_users",
                    models.ManyToManyField(
                        blank=True,
                        help_text="List of users that are allowed to see this hidden account role",
                        related_name="authorized_hidden_roles",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="authorized Hidden Users",
                    ),
                ),
                (
                    "entry",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="account_roles",
                        to="directory.entry",
                        verbose_name="Entry",
                    ),
                ),
            ],
            options={
                "verbose_name": "Account Role",
                "verbose_name_plural": "Account Roles",
            },
        ),
        migrations.CreateModel(
            name="AccountRoleType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=126, verbose_name="Title")),
                ("key", models.CharField(max_length=126, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="AccountRoleValidity",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("timespan", django.contrib.postgres.fields.ranges.DateRangeField(verbose_name="Timespan")),
                (
                    "role",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="validity_set",
                        to="wbcrm.accountrole",
                        verbose_name="Account Role",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="accountrole",
            name="role_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="roles",
                to="wbcrm.accountroletype",
                verbose_name="Role Type",
            ),
        ),
        migrations.AddConstraint(
            model_name="accountrolevalidity",
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                expressions=[("timespan", "&&"), ("role", "=")], name="exclude_overlapping_roles"
            ),
        ),
        migrations.AddConstraint(
            model_name="accountrole",
            constraint=models.UniqueConstraint(fields=("account", "entry"), name="unique_account_entry_relationship"),
        ),
    ]



from datetime import date, timedelta

from dynamic_preferences.registries import global_preferences_registry


def get_maximum_allowed_recurrent_date():
    global_preferences = global_preferences_registry.manager()
    return date.today() + timedelta(days=global_preferences["wbcrm__recurrence_maximum_allowed_days"])


def get_recurrence_maximum_count():
    global_preferences = global_preferences_registry.manager()
    return global_preferences["wbcrm__recurrence_maximum_count"]



import pytest
from django.contrib.messages import get_messages
from django.test import Client
from dynamic_preferences.registries import global_preferences_registry
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.authentication.factories import (
    InternalUserFactory,
    SuperUserFactory,
)
from wbcore.contrib.authentication.models import Permission
from wbcore.messages import InMemoryMessageStorage
from wbcore.test.utils import get_data_from_factory, get_kwargs, get_model_factory

from wbcrm.models import Activity, ActivityParticipant
from wbcrm.viewsets import (
    ActivityParticipantModelViewSet,
    ActivityTypeModelViewSet,
    ActivityViewSet,
    ProductModelViewSet,
)

# =====================================================================================================================
#                                                 TESTING ACTIVITY VIEWSETS
# =====================================================================================================================


@pytest.mark.django_db
class TestActivityModelViewSet:
    api_factory = APIRequestFactory()

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_random_cannot_see_private_item(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        item = activity_factory(visibility=CalendarItem.Visibility.PRIVATE, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == item.id
        assert response.data["results"][0]["title"] == "Private Activity"

    # =================================================================================================================
    #                                            TESTING CONFIDENTIAL CALENDAR ITEMS
    # =================================================================================================================

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_manager_can_see_confidential_item(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        permission = Permission.objects.get(codename="administrate_confidential_items")
        user.user_permissions.add(permission)
        item = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == item.id
        assert response.data["results"][0]["title"] == item.title

    @pytest.mark.parametrize("mvs", [ActivityParticipantModelViewSet])
    def test_create_activity_participant(self, mvs, activity_participant_factory):
        # Arrange
        activity_participant = activity_participant_factory()
        superuser = InternalUserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(activity_participant, mvs, delete=True, superuser=superuser)
        request = self.api_factory.post("", data=data)
        request.user = superuser
        kwargs = {"activity_id": activity_participant.activity.id}
        view = mvs.as_view({"post": "create"})
        # Act
        response = view(request, **kwargs)
        # Assert
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")
        assert response.data["instance"]["participant"] == activity_participant.participant.id
        assert (
            response.data["instance"]["participation_status"]
            == activity_participant.participation_status
            == ActivityParticipant.ParticipationStatus.PENDING_INVITATION
        )
        assert response.data["instance"]["activity"] == activity_participant.activity.id

    # @pytest.mark.parametrize("mvs", [ActivityParticipantModelViewSet])
    def test_get_activity_participant_instance(self, activity_participant_factory):
        # Arranged
        activity_participant = activity_participant_factory.create()
        user = InternalUserFactory(is_active=True, is_superuser=True)
        client = APIClient()
        url = reverse(
            "wbcrm:activity-participant-detail",
            kwargs={"pk": activity_participant.id, "activity_id": activity_participant.activity.id},
        )
        client.force_authenticate(user)
        response = client.get(url)

        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity_participant.id

    @pytest.mark.parametrize("mvs", [ActivityParticipantModelViewSet])
    def test_get_activity_participant_list(self, mvs, activity_participant_factory):
        # Arrange
        activity_participant_1 = activity_participant_factory()
        activity_participant_factory(activity=activity_participant_1.activity)
        activity_participant_factory(activity=activity_participant_1.activity)
        request = self.api_factory.get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "list"})
        kwargs = {"activity_id": activity_participant_1.activity.id}
        # Act
        response = view(request, **kwargs)
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert not response.data.get("instance")
        assert response.data.get("results")
        assert response.data["results"]
        assert len(response.data["results"]) == 3

    @pytest.mark.parametrize("mvs", [ActivityParticipantModelViewSet])
    def test_delete_activity_participant_instance(self, mvs, activity_participant_factory):
        # Arrange
        activity_participant = activity_participant_factory()
        request = self.api_factory.delete("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"delete": "destroy"})
        kwargs = {"activity_id": activity_participant.activity.id}
        # Act
        response = view(request, pk=activity_participant.id, **kwargs)
        # Assert
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize("mvs", [ActivityParticipantModelViewSet])
    def test_partial_update_activity_participant_instance(self, mvs, activity_participant_factory, person_factory):
        # Arrange
        activity_participant = activity_participant_factory()
        new_participant = person_factory()
        superuser = InternalUserFactory(is_active=True, is_superuser=True)
        request = self.api_factory.patch("", data={"participant": new_participant.id})
        request.user = superuser
        view = mvs.as_view({"patch": "partial_update"})
        kwargs = {"activity_id": activity_participant.activity.id}
        # Act
        response = view(request, pk=activity_participant.id, **kwargs)
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["participant"] == new_participant.id
        assert not response.data["instance"]["participant"] == activity_participant.participant.id

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_msg_missing_company(
        self,
        mvs,
        activity_factory,
        person_factory,
        company_factory,
    ):
        company_a = company_factory()
        company_b = company_factory()
        person = person_factory()
        person.employers.set([company_a, company_b])
        activity = activity_factory(participants=[person], companies=None, disable_participant_check=False)
        request = APIRequestFactory().get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        request.query_params = {}
        request._messages = InMemoryMessageStorage(request)
        mvs.request = request
        mvs.kwargs = get_kwargs(activity, ActivityViewSet, request)
        mvs.kwargs["pk"] = activity.pk
        mvs().add_messages(request, instance=activity)
        assert len(get_messages(request)._messages) > 0

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_msg_missing_participant(
        self,
        mvs,
        activity_factory,
        person_factory,
        company_factory,
    ):
        company = company_factory()
        person = person_factory()
        person.employers.set([company])
        company.employees.set([person])
        activity = activity_factory(participants=None, companies=[company], disable_participant_check=False)
        activity.disable_participant_check = False
        activity.save()
        request = APIRequestFactory().put("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        request.query_params = {}
        request._messages = InMemoryMessageStorage(request)
        mvs.request = request
        mvs.kwargs = get_kwargs(activity, ActivityViewSet, request)
        mvs.kwargs["pk"] = activity.pk
        mvs().add_messages(request, instance=activity)
        assert len(get_messages(request)._messages) > 0

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_msg_missing_participant_no_check(
        self,
        mvs,
        activity_factory,
        person_factory,
        company_factory,
    ):
        company = company_factory()
        person = person_factory()
        person.employers.set([company])
        company.employees.set([person])
        internal_user = InternalUserFactory.create()
        activity = activity_factory(
            participants=None, creator=internal_user.profile, companies=[company], disable_participant_check=True
        )
        request = APIRequestFactory().put("")
        request.user = internal_user
        request.query_params = {}
        request._messages = InMemoryMessageStorage(request)
        mvs.request = request
        mvs.kwargs = get_kwargs(activity, ActivityViewSet, request)
        mvs.kwargs["pk"] = activity.pk
        mvs().add_messages(request, instance=activity)
        assert len(get_messages(request)._messages) == 0

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_msg_missing_participant_and_company(
        self,
        mvs,
        activity_factory,
        person_factory,
        company_factory,
    ):
        company_a = company_factory()
        company_b = company_factory()
        company_c = company_factory()
        person_a = person_factory()
        person_b = person_factory()
        person_a.employers.set([company_a])
        person_b.employers.set([company_b, company_c])
        company_a.employees.set([person_a])
        company_b.employees.set([person_b])
        activity = activity_factory(participants=[person_b], companies=[company_a], disable_participant_check=False)
        request = APIRequestFactory().patch("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        request.query_params = {}
        request._messages = InMemoryMessageStorage(request)
        mvs.request = request
        mvs.kwargs = get_kwargs(activity, ActivityViewSet, request)
        mvs.kwargs["pk"] = activity.pk
        mvs().add_messages(request, instance=activity)
        assert len(get_messages(request)._messages) > 0

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_get_messages(self, mvs, activity_factory, recurring_activity_factory, person_factory):
        request = APIRequestFactory().get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        person1 = person_factory()
        obj = recurring_activity_factory(
            status=Activity.Status.PLANNED, repeat_choice=Activity.ReoccuranceChoice.DAILY, participants=(person1,)
        )
        activity_factory(start=obj.start, end=obj.end, participants=(person1,))
        request.query_params = {}
        request._messages = InMemoryMessageStorage(request)
        mvs.request = request
        mvs.kwargs = get_kwargs(obj, ActivityViewSet, request)
        mvs.kwargs["pk"] = obj.pk
        mvs().add_messages(request, instance=obj)
        assert len(get_messages(request)._messages) > 0

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_create_activity(self, mvs, activity_factory):
        # Arrange
        activity = activity_factory()
        user = InternalUserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(activity, mvs, delete=True, superuser=user)
        request = self.api_factory.post("", data=data)
        request.user = user
        kwargs = {}
        view = mvs.as_view({"post": "create"})
        # Act
        response = view(request, kwargs).render()
        # Assert
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")
        assert response.data["instance"]["title"] == activity.title

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_create_reviewed_activity(self, mvs, activity_factory):
        # Arrange
        activity = activity_factory(result="A Test Review For This Activity!")
        user = SuperUserFactory()
        data = get_data_from_factory(activity, mvs, delete=True, superuser=user)
        request = self.api_factory.post("", data=data)
        request.user = user
        kwargs = {}
        view = mvs.as_view({"post": "create"})
        # Act
        response = view(request, kwargs).render()
        # Assert
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")
        assert response.data["instance"]["title"] == activity.title
        assert response.data["instance"]["result"] == activity.result

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_get_activity_instance(self, mvs, activity_factory):
        # Arrange
        activity = activity_factory()
        request = self.api_factory.get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "retrieve"})
        # Act
        response = view(request, pk=activity.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_get_activity_list(self, mvs, activity_factory):
        # Arrange
        activity_factory.create_batch(3, preceded_by=None)
        request = self.api_factory.get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "list"})
        # Act
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert not response.data.get("instance")
        assert response.data.get("results")
        assert response.data["results"]
        assert len(response.data["results"]) == 3

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_delete_activity_instance(self, mvs, activity_factory):
        # Arrange
        activity = activity_factory()
        request = self.api_factory.delete("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"delete": "destroy"})
        # Act
        response = view(request, pk=activity.id).render()
        # Assert
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_update_activity_instance(self, mvs, activity_factory):
        activity_old = activity_factory()
        activity_new = activity_factory()
        user = InternalUserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(activity_new, mvs, update=True, superuser=user)
        request = self.api_factory.put("", data=data)
        request.user = user
        view = mvs.as_view({"put": "update"})
        response = view(request, pk=activity_old.id).render()
        assert response.status_code == status.HTTP_200_OK
        assert not activity_old.title == activity_new.title
        assert response.data["instance"]["id"] == activity_old.id

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_update_reviewed_activity_instance(self, mvs, activity_factory):
        user = SuperUserFactory()
        client = Client()
        client.force_login(user)
        activity: Activity = activity_factory()
        updated_result = activity.result + "Foo Bar"
        data = get_data_from_factory(activity, mvs, update=True, superuser=user)
        data["result"] = updated_result
        assert activity.result != updated_result
        update_url: str = reverse("wbcrm:activity-detail", args=[activity.pk])
        response = client.put(update_url, data, content_type="application/json")
        assert response.status_code == status.HTTP_200_OK
        activity.refresh_from_db()
        assert activity.result == updated_result

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_partial_update_activity_instance(self, mvs, activity_factory):
        activity = activity_factory()
        user = InternalUserFactory(is_active=True, is_superuser=True)
        request = self.api_factory.patch("", data={"title": "New Title"})
        request.user = user
        view = mvs.as_view({"patch": "partial_update"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["id"] == activity.id

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_partial_update_reviewed_activity_instance(self, mvs, activity_factory):
        user = SuperUserFactory()
        client = Client()
        client.force_login(user)
        activity: Activity = activity_factory()
        updated_result = activity.result + "Foo Bar"
        update_url: str = reverse("wbcrm:activity-detail", args=[activity.pk])

        assert activity.result != updated_result
        response = client.patch(update_url, {"result": updated_result}, content_type="application/json")
        assert response.status_code == status.HTTP_200_OK
        activity.refresh_from_db()
        assert activity.result == updated_result

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_conference_room_capacity_error_message(
        self, mvs, activity_factory, person_factory, conference_room_factory
    ):
        room = conference_room_factory(capacity=2)
        p1 = person_factory()
        p2 = person_factory()
        p3 = person_factory()
        activity = activity_factory(conference_room=room, participants=[p1, p2, p3])
        request = self.api_factory.get("")
        request._messages = InMemoryMessageStorage(request)
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "retrieve"})
        # Act
        response = view(request, pk=activity.id).render()
        assert len(response.data["messages"]) == 1

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_conference_room_capacity_no_error_message(
        self, mvs, activity_factory, person_factory, conference_room_factory
    ):
        room = conference_room_factory(capacity=3)
        p1 = person_factory()
        p2 = person_factory()
        p3 = person_factory()
        activity = activity_factory(conference_room=room, participants=[p1, p2, p3])
        request = self.api_factory.get("")
        request._messages = InMemoryMessageStorage(request)
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        assert len(response.data["messages"]) == 0

    # =================================================================================================================
    #                                            TESTING PRIVATE ACTIVITIES
    # =================================================================================================================

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_assignee_can_see_private_activity_list(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, assigned_to=user.profile, preceded_by=None
        )
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["title"] == activity.title
        assert response.data["results"][0]["type"] == activity.type.id
        assert response.data["results"][0]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["results"][0]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["results"][0]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["results"][0]["edited"] is not None
        assert response.data["results"][0]["created"] is not None
        assert response.data["results"][0]["description"] == activity.description
        assert response.data["results"][0]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["results"][0]["_additional_resources"] is not None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_participant_can_see_private_activity_list(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, participants=[user.profile.id], preceded_by=None
        )
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["title"] == activity.title
        assert response.data["results"][0]["type"] == activity.type.id
        assert response.data["results"][0]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["results"][0]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["results"][0]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["results"][0]["edited"] is not None
        assert response.data["results"][0]["created"] is not None
        assert response.data["results"][0]["description"] == activity.description
        assert response.data["results"][0]["result"] == activity.result
        assert response.data["results"][0]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["results"][0]["_additional_resources"] is not None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_random_cannot_see_private_activity_list(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(visibility=CalendarItem.Visibility.PRIVATE, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["title"] == "Private Activity"
        assert response.data["results"][0]["type"] is None
        assert response.data["results"][0]["edited"] is None
        assert response.data["results"][0]["created"] is None
        assert response.data["results"][0]["description"] is None
        assert response.data["results"][0]["result"] is None
        assert response.data["results"][0]["latest_reviewer"] is None
        assert response.data["results"][0]["participants"] is not None
        assert response.data["results"][0]["_participants"] is not None
        assert response.data["results"][0]["companies"] is not None
        assert response.data["results"][0]["_companies"] is not None
        assert response.data["results"][0]["groups"] is not None
        assert response.data["results"][0]["_groups"] is not None
        assert response.data["results"][0]["_additional_resources"] is None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_assignee_can_see_private_activity_instance(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, assigned_to=user.profile, preceded_by=None
        )
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id
        assert response.data["instance"]["title"] == activity.title
        assert response.data["instance"]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["instance"]["type"] == activity.type.id
        assert response.data["instance"]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["instance"]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["instance"]["edited"] is not None
        assert response.data["instance"]["created"] is not None
        assert response.data["instance"]["description"] == activity.description
        assert response.data["instance"]["result"] == activity.result
        assert response.data["instance"]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["instance"]["_additional_resources"] is not None
        assert response.data["instance"]["all_day"] == activity.all_day
        assert response.data["instance"]["assigned_to"] == activity.assigned_to.id
        assert response.data["instance"]["conference_room"] == activity.conference_room
        assert response.data["instance"]["disable_participant_check"] == activity.disable_participant_check
        assert response.data["instance"]["creator"] == activity.creator.id
        assert response.data["instance"]["importance"] == activity.importance
        assert response.data["instance"]["online_meeting"] == activity.online_meeting
        assert response.data["instance"]["reminder_choice"] == activity.reminder_choice

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_participant_can_see_private_activity_instance(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, participants=[user.profile.id], preceded_by=None
        )
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id
        assert response.data["instance"]["title"] == activity.title
        assert response.data["instance"]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["instance"]["type"] == activity.type.id
        assert response.data["instance"]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["instance"]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["instance"]["edited"] is not None
        assert response.data["instance"]["created"] is not None
        assert response.data["instance"]["description"] == activity.description
        assert response.data["instance"]["result"] == activity.result
        assert response.data["instance"]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["instance"]["_additional_resources"] is not None
        assert response.data["instance"]["all_day"] == activity.all_day
        assert response.data["instance"]["assigned_to"] == activity.assigned_to.id
        assert response.data["instance"]["conference_room"] == activity.conference_room
        assert response.data["instance"]["disable_participant_check"] == activity.disable_participant_check
        assert response.data["instance"]["creator"] == activity.creator.id
        assert response.data["instance"]["importance"] == activity.importance
        assert response.data["instance"]["online_meeting"] == activity.online_meeting
        assert response.data["instance"]["reminder_choice"] == activity.reminder_choice

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_random_cannot_see_private_activity_instance(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(visibility=CalendarItem.Visibility.PRIVATE, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id
        assert response.data["instance"]["title"] == "Private Activity"
        assert response.data["instance"]["participants"] is not None
        assert response.data["instance"]["_participants"] is not None
        assert response.data["instance"]["companies"] is not None
        assert response.data["instance"]["_companies"] is not None
        assert response.data["instance"]["groups"] is not None
        assert response.data["instance"]["_groups"] is not None
        assert response.data["instance"]["assigned_to"] is not None
        assert response.data["instance"]["_assigned_to"] is not None
        assert response.data["instance"]["creator"] is not None
        assert response.data["instance"]["_creator"] is not None
        assert response.data["instance"]["type"] is None
        assert response.data["instance"]["_type"] is None
        assert response.data["instance"]["edited"] is None
        assert response.data["instance"]["created"] is None
        assert response.data["instance"]["description"] is None
        assert response.data["instance"]["result"] is None
        assert response.data["instance"]["latest_reviewer"] is None
        assert response.data["instance"]["_latest_reviewer"] is None
        assert response.data["instance"]["_additional_resources"] is None
        assert response.data["instance"]["all_day"] is None
        assert response.data["instance"]["conference_room"] is None
        assert response.data["instance"]["disable_participant_check"] is None
        assert response.data["instance"]["importance"] is None
        assert response.data["instance"]["location"] is None
        assert response.data["instance"]["location_latitude"] is None
        assert response.data["instance"]["location_longitude"] is None
        assert response.data["instance"]["online_meeting"] is None
        assert response.data["instance"]["reminder_choice"] is None
        assert response.data["instance"]["reviewed_at"] is None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_private_instance_not_deleteable(self, mvs, activity_factory, user_factory):
        activity = activity_factory(visibility=CalendarItem.Visibility.PRIVATE, preceded_by=None)
        request = self.api_factory.delete("")
        request.user = user_factory(is_active=True, is_superuser=True)
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_private_instance_deleteable_for_participant(self, mvs, activity_factory, user_factory):
        superuser = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, participants=[superuser.profile.id], preceded_by=None
        )
        request = self.api_factory.delete("")
        request.user = superuser
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_private_instance_deleteable_for_assignee(self, mvs, activity_factory, user_factory):
        superuser = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(
            visibility=CalendarItem.Visibility.PRIVATE, assigned_to=superuser.profile, preceded_by=None
        )
        request = self.api_factory.delete("")
        request.user = superuser
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_204_NO_CONTENT

    # =================================================================================================================
    #                                            TESTING CONFIDENTIAL ACTIVITIES
    # =================================================================================================================

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_manager_can_see_confidential_activity_list(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        permission = Permission.objects.get(codename="administrate_confidential_items")
        user.user_permissions.add(permission)
        activity = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["title"] == activity.title
        assert response.data["results"][0]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["results"][0]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["results"][0]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["results"][0]["type"] == activity.type.id
        assert response.data["results"][0]["edited"] is not None
        assert response.data["results"][0]["created"] is not None
        assert response.data["results"][0]["description"] == activity.description
        assert response.data["results"][0]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["results"][0]["_additional_resources"] is not None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_random_cannot_see_confidential_activity_list(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["title"] == "Confidential Activity"
        assert response.data["results"][0]["participants"] is None
        assert response.data["results"][0]["companies"] is None
        assert response.data["results"][0]["groups"] is None
        assert response.data["results"][0]["type"] is None
        assert response.data["results"][0]["edited"] is None
        assert response.data["results"][0]["created"] is None
        assert response.data["results"][0]["description"] is None
        assert response.data["results"][0]["result"] is None
        assert response.data["results"][0]["latest_reviewer"] is None
        assert response.data["results"][0]["_additional_resources"] is None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_manager_can_see_confidential_activity_instance(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        permission = Permission.objects.get(codename="administrate_confidential_items")
        user.user_permissions.add(permission)
        activity = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id
        assert response.data["instance"]["title"] == activity.title
        assert response.data["instance"]["participants"] == list(
            activity.participants.all().values_list("id", flat=True)
        )
        assert response.data["instance"]["type"] == activity.type.id
        assert response.data["instance"]["companies"] == list(activity.companies.all().values_list("id", flat=True))
        assert response.data["instance"]["groups"] == list(activity.groups.all().values_list("id", flat=True))
        assert response.data["instance"]["edited"] is not None
        assert response.data["instance"]["created"] is not None
        assert response.data["instance"]["description"] == activity.description
        assert response.data["instance"]["result"] == activity.result
        assert response.data["instance"]["latest_reviewer"] == activity.latest_reviewer.id
        if activity.status not in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
            assert response.data["instance"]["_additional_resources"] is not None
        assert response.data["instance"]["all_day"] == activity.all_day
        assert response.data["instance"]["assigned_to"] == activity.assigned_to.id
        assert response.data["instance"]["conference_room"] == activity.conference_room
        assert response.data["instance"]["disable_participant_check"] == activity.disable_participant_check
        assert response.data["instance"]["creator"] == activity.creator.id
        assert response.data["instance"]["importance"] == activity.importance
        assert response.data["instance"]["online_meeting"] == activity.online_meeting
        assert response.data["instance"]["reminder_choice"] == activity.reminder_choice

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_random_cannot_see_confidential_activity_instance(self, mvs, activity_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "retrieve"})
        response = view(request, pk=activity.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == activity.id
        assert response.data["instance"]["title"] == "Confidential Activity"
        assert response.data["instance"]["participants"] is None
        assert response.data["instance"]["_participants"] is None
        assert response.data["instance"]["type"] is None
        assert response.data["instance"]["_type"] is None
        assert response.data["instance"]["companies"] is None
        assert response.data["instance"]["_companies"] is None
        assert response.data["instance"]["groups"] is None
        assert response.data["instance"]["_groups"] is None
        assert response.data["instance"]["edited"] is None
        assert response.data["instance"]["created"] is None
        assert response.data["instance"]["description"] is None
        assert response.data["instance"]["result"] is None
        assert response.data["instance"]["latest_reviewer"] is None
        assert response.data["instance"]["_latest_reviewer"] is None
        assert response.data["instance"]["_additional_resources"] is None
        assert response.data["instance"]["all_day"] is None
        assert response.data["instance"]["assigned_to"] is None
        assert response.data["instance"]["conference_room"] is None
        assert response.data["instance"]["disable_participant_check"] is None
        assert response.data["instance"]["creator"] is None
        assert response.data["instance"]["importance"] is None
        assert response.data["instance"]["location"] is None
        assert response.data["instance"]["location_latitude"] is None
        assert response.data["instance"]["location_longitude"] is None
        assert response.data["instance"]["online_meeting"] is None
        assert response.data["instance"]["reminder_choice"] is None
        assert response.data["instance"]["reviewed_at"] is None
        assert response.data["instance"]["_assigned_to"] is None
        assert response.data["instance"]["_creator"] is None

    @pytest.mark.parametrize("mvs", [ActivityViewSet])
    def test_confidential_instance_not_deleteable(self, mvs, activity_factory, internal_user_factory, user_factory):
        employee = internal_user_factory().profile
        user = user_factory(is_active=True)
        employee.user_account = user
        for permission in Permission.objects.exclude(codename="administrate_confidential_items"):
            user.user_permissions.add(permission)
        global_preferences_registry.manager()["directory__main_company"] = employee.employers.first().id
        activity = activity_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL, preceded_by=None)
        request = self.api_factory.delete("")
        request.user = user
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=activity.id).render()
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED


# =====================================================================================================================
#                                                  TESTING ENTRY VIEWSETS
# =====================================================================================================================


# # =====================================================================================================================
# #                                                  TESTING UTILS VIEWSETS
# # =====================================================================================================================


@pytest.mark.django_db
class TestUtilsViewSets:
    api_factory = APIRequestFactory()

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_get_utils(self, mvs):
        request = self.api_factory.get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        factory.create_batch(3)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_retrieve_utils(self, mvs):
        request = self.api_factory.get("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_post_utils(self, mvs):
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        super_user = InternalUserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(obj, mvs, superuser=super_user, delete=True)
        request = self.api_factory.post("", data=data)
        request.user = super_user
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_delete_utils(self, mvs):
        request = self.api_factory.delete("")
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_put_utils(self, mvs):
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        old_obj = factory()
        new_obj = factory()
        user = InternalUserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(new_obj, mvs, superuser=user, delete=True)
        request = APIRequestFactory().put("", data=data)
        request.user = user
        vs = mvs.as_view({"put": "update"})
        response = vs(request, pk=old_obj.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == new_obj.title
        assert not response.data["instance"]["title"] == old_obj.title

    @pytest.mark.parametrize(
        "mvs",
        [
            ActivityTypeModelViewSet,
            ProductModelViewSet,
        ],
    )
    def test_patch_utils(self, mvs):
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        request = APIRequestFactory().patch("", data={"title": "New Title"})
        request.user = InternalUserFactory(is_active=True, is_superuser=True)
        vs = mvs.as_view({"patch": "partial_update"})
        response = vs(request, pk=obj.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == "New Title"



from django.db.models import QuerySet
from dynamic_preferences.registries import global_preferences_registry
from googleapiclient.discovery import Resource
from wbcrm.models import Activity


def cancel_or_delete_activity(activity: "Activity") -> None:
    # Activity is cancelled rather than disable if global preference is True
    if global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_activity"]:
        Activity.objects.filter(id=activity.id).update(status=Activity.Status.CANCELLED)
    else:
        activity.delete()


def cancel_or_delete_activity_queryset(activity_qs: QuerySet["Activity"]) -> None:
    # Activities are cancelled rather than delete if global preference is True
    if global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_activity"]:
        activity_qs.update(status=Activity.Status.CANCELLED)
    else:
        activity_qs.delete()


def delete_single_activity(activity: Activity):
    """
    Deletes the activity that has the corresponding external ID
    """
    Activity.objects.filter(id=activity.id).update(external_id=None)
    cancel_or_delete_activity(activity)


def delete_recurring_activity(activity: Activity, event: dict, user_mail: str, service: Resource):
    """
    Handles the deletion of recurring activities (either a single activity, a certain number of activities or all activities),
    based on the changes done to the google event.

    :param event:           The google event dict that was deleted.
    :param activity:        The corresponding workbench activity.
    :param user_mail:       The e-mail address of the current user.
    :param service:         The google service to interact with googles resources.
    """

    parent_occurrence = activity.parent_occurrence if activity.parent_occurrence else activity
    metadata_event = parent_occurrence.metadata["google_backend"].get(
        "event", parent_occurrence.metadata["google_backend"].get("instance", {})
    )
    if event.get("recurringEventId"):
        # Delete single event in event chain
        if activity.status not in [Activity.Status.REVIEWED, Activity.Status.FINISHED]:
            cancel_or_delete_activity(activity)

    elif event.get("recurrence") != metadata_event.get("recurrence"):
        # Delete all events after a certain event in the event chain

        event_instances: dict = service.events().instances(calendarId=user_mail, eventId=event["id"]).execute()
        external_id_list: list = [instance["id"] for instance in event_instances["items"]]
        activities_to_remove = Activity.objects.filter(
            metadata__google_backend__instance__recurringEventId__startswith=event["id"],
            status=Activity.Status.PLANNED,
        ).exclude(metadata__google_backend__instance__id__in=external_id_list)
        first_in_list = activities_to_remove.order_by("start").first()
        metadata = parent_occurrence.metadata
        metadata["google_backend"] |= {"event": event}
        Activity.objects.filter(id=parent_occurrence.id).update(metadata=metadata)
        if (
            first_in_list
            and first_in_list.id == parent_occurrence.id
            and parent_occurrence.status not in [Activity.Status.REVIEWED, Activity.Status.FINISHED]
        ):
            cancel_or_delete_activity(parent_occurrence)
        cancel_or_delete_activity_queryset(activities_to_remove)

    else:
        # Delete all events in event chain
        activities_to_cancel = Activity.objects.filter(
            parent_occurrence=parent_occurrence, status=Activity.Status.PLANNED
        )
        cancel_or_delete_activity_queryset(activities_to_cancel)
        cancel_or_delete_activity(activity)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ProductCompanyRelationshipEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcrm:company-interestedproduct-list", args=[self.view.kwargs["company_id"]], request=self.request
        )



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

GROUPS_MENUITEM = MenuItem(
    label=_("Groups"),
    endpoint="wbcrm:group-list",
    permission=ItemPermission(method=lambda request: is_internal_user(request.user), permissions=["wbcrm.view_group"]),
    add=MenuItem(
        label=_("Create Groups"),
        endpoint="wbcrm:group-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user), permissions=["wbcrm.add_group"]
        ),
    ),
)



# Generated by Django 4.2.6 on 2024-01-31 08:30

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0006_alter_activity_location"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="status",
            field=django_fsm.FSMField(
                choices=[("PENDING", "Pending"), ("OPEN", "Open"), ("CLOSE", "Close")],
                default="OPEN",
                max_length=50,
                verbose_name="Status",
            ),
        ),
    ]
